local LMP = LibMapPins

local DEG_ADDON = _G["DEG_CURRENT_ADDON"]
local function d(msg)
  _G[DEG_ADDON.PACKAGE_NAME].plugins[DEG_ADDON.ADDON_NAME_SHORT]:d(msg)
end

local function ts(...)
  return tostring(...)
end

local Addon = {}
Addon.initialized = false
Addon.debug = false--release:false
Addon.name = DEG_ADDON.ADDON_NAME
Addon.versionString = '1.12'
Addon.saveVariablesName = DEG_ADDON.SAVED_VARS_NAME
Addon.savedVariablesAccount = nil
Addon.savedVariablesCharacter = nil
Addon.saveVariablesVersion = 1
Addon.vars = {control = nil}
Addon.Settings = _G[DEG_ADDON.ADDON_NAME.."Settings"]

function Addon:Initialize()
  if (self.initialized) then return end

  local defaultsAccount = {}
  self.savedVariablesAccount = ZO_SavedVars:NewAccountWide(self.saveVariablesName, self.saveVariablesVersion, nil, defaultsAccount)

  local defaultsCharacter = {}
  self.savedVariablesCharacter = ZO_SavedVars:New(self.saveVariablesName, self.saveVariablesVersion, nil, defaultsCharacter)

  self.Settings:initialize()

  self.worldControlPool = ZO_ControlPool:New("DEGLorePin", DEGLorePins, "DEGLorePin")
  Lib3D:RegisterWorldChangeCallback("DEGLorePin", function(...) self:onWorldChange(...) end)

  EVENT_MANAGER:RegisterForEvent(self.name, EVENT_PLAYER_ACTIVATED, function(...) self:onEVENT_PLAYER_ACTIVATED(...) end)
  EVENT_MANAGER:RegisterForEvent(self.name, EVENT_PLAYER_DEACTIVATED, function(...) self:onEVENT_PLAYER_DEACTIVATED(...) end)
  EVENT_MANAGER:RegisterForEvent(self.name, EVENT_ZONE_CHANGED, function(...) self:onEVENT_ZONE_CHANGED(...) end)

  CALLBACK_MANAGER:RegisterCallback("OnWorldMapChanged", function(...) self:onWorldMapChange(...) end)

  self.initialized = true
end

function Addon:CharWantsBooks(charId)
  if not charId then
    charId = GetCurrentCharacterId()
  end
  if self.savedVariablesAccount['charWantsBooks'] then
    if self.savedVariablesAccount['charWantsBooks'][charId] ~= nil then
      return self.savedVariablesAccount['charWantsBooks'][charId]
    end
  end
  return true
end

function Addon:setCharWantsBooks(charId, newValue)
  if not self.savedVariablesAccount['charWantsBooks'] then
    self.savedVariablesAccount['charWantsBooks'] = {}
    if not self.savedVariablesAccount['charWantsBooks'][charId] then
      self.savedVariablesAccount['charWantsBooks'][charId] = true
    end
  end
  self.savedVariablesAccount['charWantsBooks'][charId] = newValue

  if (newValue) then
    self:reinitPins()
  else

  end
  self:checkRefreshHandler()
end

function Addon:checkRefreshHandler()
  if self:CharWantsBooks() then
    DEGLorePins:SetHidden(false)
    self:reinitPins()
    EVENT_MANAGER:RegisterForUpdate("DryzlerElderGeekLoreUpdate", 200, function(...) self:onUpdate(...)  end)
  else
    DEGLorePins:SetHidden(true)
    EVENT_MANAGER:UnregisterForUpdate("DryzlerElderGeekLoreUpdate")
  end
end

Addon.pins= {

}

function Addon:onWorldMapChange()
  d("onWorldMapChange")

  self:reinitPins()
end

function Addon:onWorldChange(identifier, zoneIndex, validZone, newWorld)
  d("onWorldChange;"..ts(identifier)..";"..ts(zoneIndex)..";"..ts(validZone)..";" .. ts(newWorld))
  --Addon:onWorldChange;DEGLorePin;178;true;true

  if not Addon:CharWantsBooks(GetCurrentCharacterId()) then return end

  if not validZone then return end

  self:reinitPins()
end

--CALLBACK_MANGER:FireCallbacks("OnWorldMapChanged")

--    { 0.5928, 0.8493, 19, 4 },    -- History of the Fighter's Guilds Pt.2
--    { 0.6196, 0.8249, 19, 4 },
--    { 0.5993, 0.7961, 19, 4 },
--    { 0.5679, 0.8275, 19, 4 },

Addon.currentZone = nil

function Addon:reinitPins()
  d("Addon:reinitPins")


  if not Lib3D.currentZoneMeasurement then
    d("Skipping reinitPins, not Lib3D.currentZoneMeasurement")
    return
  end
--  SetMapToPlayerLocation()

  if not DEGLorePins:Has3DRenderSpace()  then
    DEGLorePins:Create3DRenderSpace()
  end

  DEGLorePins:Set3DRenderSpaceOrigin(0,0,0)

  --local zoneIndex = GetUnitZoneIndex("player")
  --if IsValidZone(zoneIndex) then
    local zone, subzone = LMP:GetZoneAndSubzone()
    if zone and subzone then

      local myZoneId  = ts(zone)..";"..ts(subzone)

      if (self.currentZone and self.currentZone ~= myZoneId) then
        self.worldControlPool:ReleaseAllObjects()
        self.pins = {}
        self.currentZone = myZoneId
      end

      self.worldControlPool:ReleaseAllObjects()

      local lorebooks = LoreBooks_GetLocalData(GetCurrentMapId())
      if lorebooks then
        --d("Addon:reinitPins3")
        for pinIndex, pinData in ipairs(lorebooks) do
          local _, _, known = GetLoreBookInfo(1, pinData[3], pinData[4])

          if not known then
            local myPinKey = ts(zone..";"..ts(subzone)..";"..pinIndex)
            local pinKey = self.pins[myPinKey]


            local pin
            if (not pinKey) then
              pin, pinKey = self.worldControlPool:AcquireObject()
            else
              pin = self.worldControlPool:GetExistingObject(pinKey)
              if (not pin) then
                pin = self.worldControlPool:AcquireObject()
              else
                --return
              end
            end
            pin:SetHidden(true)

            pin.myPinKey = myPinKey

            self.pins[myPinKey] = pinKey

            local x,y = Lib3D:LocalToWorld(pinData[1], pinData[2])
  --            d("x="..tostring(x)..";y="..tostring(y)..";z="..tostring(z))
            x=x*100
            y=y*100

            pin.x = x
            pin.y = y

            pin.pinData = pinData

            local beam = pin:GetNamedChild("Beam")
            pin.beam = beam
            local icon = pin:GetNamedChild("Icon")
            pin.icon = icon
            if not pin:Has3DRenderSpace()  then
              pin:Create3DRenderSpace()
          --    beam:Create3DRenderSpace()
          --    beam:SetTexture("HarvestMap/Textures/worldMarker.dds")
              icon:Create3DRenderSpace()
              icon:SetTexture("Lorebooks/Icons/book1.dds")
            end

          --  pin:Set3DRenderSpaceUsesDepthBuffer(false)
          --  beam:Set3DRenderSpaceUsesDepthBuffer(false)
          --  icon:Set3DRenderSpaceUsesDepthBuffer(false)

            local width = 1
            local height = 2

            icon:Set3DRenderSpaceOrigin(0, height, 0)
            --icon:Set3DRenderSpaceOrigin(0, height + 0.125 * height + 0.25, 0)
            icon:Set3DLocalDimensions(0.25 * height + 0.5, 0.25 * height + 0.5)

            local _,x,z,y = GetUnitWorldPosition("player") --centimeters
            local worldX, worldY, worldZ = WorldPositionToGuiRender3DPosition(pin.x, z, pin.y)
            pin:Set3DRenderSpaceOrigin(worldX, worldY, worldZ)
          end
        end
      end
    end
  --end
end

function Addon:onUpdate()
--  if true then return end
  if self.deactivated then return end

  if not Addon:CharWantsBooks(GetCurrentCharacterId()) then return end

--  d("onUpdate")

  local zoneTemp,x,z,y = GetUnitWorldPosition("player") --centimeters

  --d("GetUnitWorldPosition;zoneTemp="..ts(zoneTemp)..";x="..ts(x)..";z="..ts(z).."y="..ts(y))
  --draußen
  --GetUnitWorldPosition;zoneTemp=381;x=231176;z=12025y=379500
  --drinnen
  --GetUnitWorldPosition;zoneTemp=381;x=231523;z=12040y=380202

  local heading = GetPlayerCameraHeading()
  if heading > math.pi then --normalize heading to [-pi,pi]
    heading = heading - 2 * math.pi
  end

  --spieler
  local Ax = x
  local Ay = y

  for i, control in pairs(self.worldControlPool.m_Active) do
    control:Set3DRenderSpaceOrientation(0, heading, 0)

    if true then
--      local x,y = Lib3D:LocalToWorld(control.pinData[1], control.pinData[2])
--      x=x*100
--      y=y*100

--      control.x = x
--      control.y = y

      local Bx = control.x
      local By = control.y

      local Cx = Ax
      local Cy = By

      --d("By="..ts(By)..";Ay="..ts(Ay))

      local Dy = By - Ay
      local Dx = Bx - Ax

      local dist = math.sqrt((Dx * Dx) + (Dy * Dy)) / 100  --meter

      if (dist <= 50) then --50 meter
        local _, _, known = GetLoreBookInfo(1, control.pinData[3], control.pinData[4])

        if (known) then
          control:SetHidden(true)
        else
          --d("activating:pin.myPinKey="..ts(control.myPinKey))
          control:SetHidden(false)

          local worldX, worldY, worldZ = WorldPositionToGuiRender3DPosition(control.x, z, control.y)
          control:Set3DRenderSpaceOrigin(worldX, worldY, worldZ)

          --if (dist <= 4) then
            --control.icon:SetColor(0.0*255/255,0.5*255/255,0.0*255/255,1*255/255)
            --control.hasColor = true
          --else
            --if (control.hasColor) then
              --control.icon:SetColor(1,1,1)
              --control.hasColor = false
            --end
          --end
        end
      else
        control:SetHidden(true)
      end
    end
  end
end

function Addon:onEVENT_ZONE_CHANGED (eventCode, zoneName, subZoneName, bNewSubzone, zoneId, subZoneId)
  d("onEVENT_ZONE_CHANGED")

  self:reinitPins()
end

function Addon:onEVENT_PLAYER_DEACTIVATED(intEventCode)
  d("onEVENT_PLAYER_DEACTIVATED: "..GetUnitName("player"))

  self.deactivated = true
  EVENT_MANAGER:UnregisterForUpdate("DryzlerElderGeekLoreUpdate")
end

function Addon:onEVENT_PLAYER_ACTIVATED(intEventCode, bInitial)
  d("onEVENT_PLAYER_ACTIVATED: "..GetUnitName("player")..";bInitial="..ts(bInitial))

  self.deactivated = false
  --self:reinitPins()

  self:checkRefreshHandler();



--  EVENT_MANAGER:RegisterForUpdate("DryzlerElderGeekLoreUpdateDebug", 5000, function(...)
--    self:reinitPins()
--  end)

end

--#################################################################################################

function Addon:d(m)
  if self.debug then
    _G.d(self.name.."> "..tostring(m))
  end
end

_G[DEG_ADDON.PACKAGE_NAME].plugins[DEG_ADDON.ADDON_NAME_SHORT] = Addon;

EVENT_MANAGER:RegisterForEvent(DEG_ADDON.ADDON_NAME, EVENT_ADD_ON_LOADED,
  function(event, AddonName)
    if AddonName == _G[DEG_ADDON.PACKAGE_NAME].plugins[DEG_ADDON.ADDON_NAME_SHORT].name then
      _G[DEG_ADDON.PACKAGE_NAME].plugins[DEG_ADDON.ADDON_NAME_SHORT]:Initialize()
      EVENT_MANAGER:UnregisterForEvent(_G[DEG_ADDON.PACKAGE_NAME].plugins[DEG_ADDON.ADDON_NAME_SHORT].name, EVENT_ADD_ON_LOADED)
    end
  end
)